#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define PAYLOAD_SIZE_OFFSET 0x4a4c
#define PAYLOAD_START_OFFSET 0xa3fc
#define PAYLOAD_BLOCK_HEADER_SIZE 0x24
#define PAYLOAD_BLOCK_BODY_SIZE 0x100
#define PAYLOAD_ZEROBLOCK_SIZE 0xdc
#define PAYLOAD_BLOCK_SIZE (PAYLOAD_BLOCK_HEADER_SIZE + PAYLOAD_BLOCK_BODY_SIZE + PAYLOAD_ZEROBLOCK_SIZE)
#define PAYLOAD_HEAD_PADDING_SIZE 0xa8
#define PAYLOAD_TAIL_INTERMEZZO_SIZE 92
#define PAYLOAD_TAIL_ZEROPAGE_COUNT 5
#define PAYLOAD_TAIL_PADDINGPAGE_COUNT 5

const char intermezzo[PAYLOAD_TAIL_INTERMEZZO_SIZE] =
{
    0x44, 0x00, 0x9F, 0xE5, 0x01, 0x11, 0xA0, 0xE3, 0x40, 0x20, 0x9F, 0xE5, 0x00, 0x20, 0x42, 0xE0, 
    0x08, 0x00, 0x00, 0xEB, 0x01, 0x01, 0xA0, 0xE3, 0x10, 0xFF, 0x2F, 0xE1, 0x00, 0x00, 0xA0, 0xE1, 
    0x2C, 0x00, 0x9F, 0xE5, 0x2C, 0x10, 0x9F, 0xE5, 0x02, 0x28, 0xA0, 0xE3, 0x01, 0x00, 0x00, 0xEB, 
    0x20, 0x00, 0x9F, 0xE5, 0x10, 0xFF, 0x2F, 0xE1, 0x04, 0x30, 0x90, 0xE4, 0x04, 0x30, 0x81, 0xE4, 
    0x04, 0x20, 0x52, 0xE2, 0xFB, 0xFF, 0xFF, 0x1A, 0x1E, 0xFF, 0x2F, 0xE1, 0x20, 0xF0, 0x01, 0x40, 
    0x5C, 0xF0, 0x01, 0x40, 0x00, 0x00, 0x02, 0x40, 0x00, 0x00, 0x01, 0x40,
};

void die(int errCode);
int patchFirmware(FILE *fw, FILE *payload);
long getFileSize(FILE *f);
long getOldPayloadSize(FILE *fw);
int patchPayloadSize(FILE *fw, long payloadSize);
int patchIntVal(FILE *fw, long offset, long intVal, int numbytes);
int patchShort(FILE *fw, long offset, short shortVal);
int patchPayloadBytes(FILE *fw, FILE *payload, long newPayloadSize, long oldPayloadSize, int lastPatchVal);
int patchFirstPayloadPage(FILE *fw, FILE *payload, long firstPageSize);
int copyPayloadTailPages(FILE *fw, long oldPayloadNumPages, long newPayloadNumPages, int lastPatchVal);
int writePages(FILE *fw, char *pages, int pageCount, long offset);
int patchTailEmptyPages(FILE *fw, long payloadNumPages);


int main(int argc, char **argv) {
  if (argc != 3) {
    die(1);
  }
  int errCode = 0;

  FILE *fw = fopen(argv[1], "rb+");
  if (fw != NULL) {
    FILE *payload = fopen(argv[2], "rb");
    if (payload != NULL) {
      errCode = patchFirmware(fw, payload);
      fclose(payload);
    } else {
      errCode = 3;
    }
    fclose(fw);

  } else {
    errCode = 2;
  }

  if (errCode) {
    die(errCode);
  }
  return 0;
}

void die(int errCode) {
  /*
    stub
    TODO: error handling/display
  */
  fprintf(stderr, "Error code: [%i]\n", errCode);
  exit(errCode);
}

#define tryCatch(e,x) if((x)==-1){return e;}

int patchFirmware(FILE *fw, FILE *payload){
  long payloadSize = getFileSize(payload);
  if (payloadSize == -1) {
    return 4;
  }
  if (payloadSize > 0x30000) {
    return 5;
  }
  long oldPayloadSize;
  tryCatch(6, oldPayloadSize = getOldPayloadSize(fw))
  tryCatch(7, patchPayloadSize(fw, payloadSize));
  long patchValL = payloadSize + 0x52a8;
  tryCatch(8, patchIntVal(fw, 0x4698, patchValL, 3));
  int sizeUpperBits = payloadSize >> 8;
  int patchVal = sizeUpperBits + 0x5c;
  tryCatch(9, patchShort(fw, 0x40f9, patchVal));
  tryCatch(10, patchShort(fw, 0x4225, patchVal));
  tryCatch(11, patchShort(fw, 0x60c5, patchVal));
  patchVal = sizeUpperBits + 0x54;
  tryCatch(12, patchShort(fw, 0x4a7d, patchVal));
  patchVal = sizeUpperBits + 0x59;
  tryCatch(13, patchShort(fw, 0x4abd, patchVal));
  tryCatch(14, patchShort(fw, 0x4b05, patchVal));
  tryCatch(15, patchShort(fw, 0x68c1, patchVal));
  tryCatch(16, patchShort(fw, 0x6a21, patchVal));
  tryCatch(17, patchShort(fw, 0x6ae9, patchVal));
  tryCatch(18, patchShort(fw, 0x7af1, patchVal));
  patchVal = sizeUpperBits + 0x5B;
  tryCatch(19, patchShort(fw, 0x78e1, patchVal));
  tryCatch(20, patchShort(fw, 0x7e81, patchVal));
  tryCatch(21, patchShort(fw, 0x8b19, patchVal));
  tryCatch(22, patchShort(fw, 0x96e9, patchVal));
  tryCatch(23, patchShort(fw, 0x96ed, patchVal));
  tryCatch(24, patchShort(fw, 0x96f1, patchVal));
  tryCatch(25, patchShort(fw, 0x96f5, patchVal));

  return patchPayloadBytes(fw, payload, payloadSize, oldPayloadSize, patchVal);
}

long getFileSize(FILE *f) {
  if (fseek(f, 0, SEEK_END)) {
    return -1;
  }
  return ftell(f);
}

long getOldPayloadSize(FILE *fw) {
  if (fseek(fw, PAYLOAD_SIZE_OFFSET, SEEK_SET)) {
    return -1;
  }
  unsigned char sizeBytes[3];
  if (fread(sizeBytes, sizeof(char), 3, fw) == 3) {
    return sizeBytes[0] + (sizeBytes[1] << 8) + (sizeBytes[2] << 16);
  }
  return -1;
}

int patchPayloadSize(FILE *fw, long payloadSize) {
  return patchIntVal(fw, PAYLOAD_SIZE_OFFSET, payloadSize, 3);
}

int patchIntVal(FILE *fw, long offset, long intVal, int numbytes) {
  if (fseek(fw, offset, SEEK_SET)) {
    return -1;
  }
  char sizeBytes[numbytes];
  int i;
  for (i=0; i<numbytes; i++) {
    sizeBytes[i] = intVal & 0xff;
    intVal >>= 8;
  }
  if (fwrite(sizeBytes, sizeof(char), numbytes, fw) != numbytes) {
    return -1;
  }
  return 0;
}

int patchShort(FILE *fw, long offset, short shortVal) {
  return patchIntVal(fw, offset, shortVal, 2);
}

#define ceilDiv(x,y) (((x) + (y) - 1)/(y))
#define getPageCount(payloadSizeVar) (ceilDiv(payloadSizeVar + PAYLOAD_HEAD_PADDING_SIZE + PAYLOAD_TAIL_INTERMEZZO_SIZE, PAYLOAD_BLOCK_BODY_SIZE))

int patchPayloadBytes(FILE *fw, FILE *payload, long newPayloadSize, long oldPayloadSize, int lastPatchVal) {
  long oldPayloadNumPages = getPageCount(oldPayloadSize);
  long newPayloadNumPages = getPageCount(newPayloadSize);

  tryCatch(26, copyPayloadTailPages(fw, oldPayloadNumPages, newPayloadNumPages, lastPatchVal));
  tryCatch(27, patchTailEmptyPages(fw, newPayloadNumPages));

  long firstPageSize = PAYLOAD_BLOCK_BODY_SIZE - PAYLOAD_HEAD_PADDING_SIZE;
  tryCatch(28, patchFirstPayloadPage(fw, payload, firstPageSize));

  long remainingPayload = newPayloadSize - firstPageSize;
  long payloadBytesToWriteLeft = remainingPayload + PAYLOAD_TAIL_INTERMEZZO_SIZE;
  long pagesLeft = getPageCount(payloadBytesToWriteLeft);
  long payloadPagesByteCount = pagesLeft * PAYLOAD_BLOCK_BODY_SIZE;
  long zeroBytes = payloadPagesByteCount - payloadBytesToWriteLeft;
  char *payloadBuffer = malloc(sizeof(char) * payloadPagesByteCount);
  if (payloadBuffer == NULL) {
    return 29;
  }
  if (fseek(payload, firstPageSize, SEEK_SET) != 0) {
    free(payloadBuffer);
    return 30;
  }
  int readBytes = fread(payloadBuffer, sizeof(char), remainingPayload, payload);
  if (readBytes != remainingPayload) {
    free(payloadBuffer);
    return 30;
  }
  memcpy(payloadBuffer + remainingPayload, intermezzo, PAYLOAD_TAIL_INTERMEZZO_SIZE);
  memset(payloadBuffer + payloadBytesToWriteLeft, 0, zeroBytes);
  long pagesLeftOffset = PAYLOAD_START_OFFSET + PAYLOAD_BLOCK_SIZE;
  writePages(fw, payloadBuffer, pagesLeft, pagesLeftOffset);
  
  free(payloadBuffer);
  return 0;
}

int patchFirstPayloadPage(FILE *fw, FILE *payload, long firstPageSize) {
  char *buffer = malloc(sizeof(char) * firstPageSize);
  if (buffer != NULL) {
    long firstPageOffset = PAYLOAD_START_OFFSET + PAYLOAD_BLOCK_HEADER_SIZE + PAYLOAD_HEAD_PADDING_SIZE;
    rewind(payload);
    int readBytes = fread(buffer, sizeof(char), firstPageSize, payload);
    if (readBytes == firstPageSize) {
      if (fseek(fw, firstPageOffset, SEEK_SET) == 0) {
        int writtenBytes = fwrite(buffer, sizeof(char), firstPageSize, fw);
        if (writtenBytes == firstPageSize) {
          free(buffer);
          return 0;
        }
      }
    }
    free(buffer);
  }
  return -1;
}

int copyPayloadTailPages(FILE *fw, long oldPayloadNumPages, long newPayloadNumPages, int lastPatchVal) {
  long oldTailPagesOffset = PAYLOAD_START_OFFSET + ((oldPayloadNumPages + PAYLOAD_TAIL_ZEROPAGE_COUNT) * PAYLOAD_BLOCK_SIZE);
  char *tailPagesCopy = malloc(sizeof(char) * PAYLOAD_TAIL_PADDINGPAGE_COUNT * PAYLOAD_BLOCK_BODY_SIZE);
  if (tailPagesCopy == NULL) {
    return -1;
  }
  int i;
  for (i=0; i < PAYLOAD_TAIL_PADDINGPAGE_COUNT; i++) {
    long curOffset = oldTailPagesOffset + (i * PAYLOAD_BLOCK_SIZE) + PAYLOAD_BLOCK_HEADER_SIZE;
    if (fseek(fw, curOffset, SEEK_SET) == 0) {
      int readBytes = fread(tailPagesCopy + (i * PAYLOAD_BLOCK_BODY_SIZE), sizeof(char), PAYLOAD_BLOCK_BODY_SIZE, fw);
      if (readBytes == PAYLOAD_BLOCK_BODY_SIZE) {
        continue;
      }
    }
    free(tailPagesCopy);
    return -1;
  }
  int patchValOffset = (3 * PAYLOAD_BLOCK_BODY_SIZE) + 0xb1;
  tailPagesCopy[patchValOffset] = lastPatchVal & 0xff;
  tailPagesCopy[patchValOffset+1] = (lastPatchVal >> 8) & 0xff;
  
  long newTailPagesOffset = PAYLOAD_START_OFFSET + ((newPayloadNumPages + PAYLOAD_TAIL_ZEROPAGE_COUNT) * PAYLOAD_BLOCK_SIZE);
  int retVal = writePages(fw, tailPagesCopy, PAYLOAD_TAIL_PADDINGPAGE_COUNT, newTailPagesOffset);
  free(tailPagesCopy);
  return retVal;
}

int writePages(FILE *fw, char *pages, int pageCount, long offset) {
  int i;
  for (i=0; i<pageCount; i++) {
    long curOffset = offset + (i * PAYLOAD_BLOCK_SIZE) + PAYLOAD_BLOCK_HEADER_SIZE;
    if (fseek(fw, curOffset, SEEK_SET) != 0) {
      return -1;
    }
    int writtenBytes = fwrite(pages + (i * PAYLOAD_BLOCK_BODY_SIZE), sizeof(char), PAYLOAD_BLOCK_BODY_SIZE, fw);
    if (writtenBytes != PAYLOAD_BLOCK_BODY_SIZE) {
      return -1;
    }
  }
  return 0;
}

int patchTailEmptyPages(FILE *fw, long payloadNumPages){
  char *dummyPage = malloc(sizeof(char) * PAYLOAD_BLOCK_BODY_SIZE);
  if (dummyPage == NULL) {
    return -1;
  }
  memset(dummyPage, 0, sizeof(char) * PAYLOAD_BLOCK_BODY_SIZE);
  long zeroPagesOffset = PAYLOAD_START_OFFSET + (payloadNumPages * PAYLOAD_BLOCK_SIZE);
  int i;
  for (i=0; i<PAYLOAD_TAIL_ZEROPAGE_COUNT; i++) {
    long curOffset = zeroPagesOffset + (i * PAYLOAD_BLOCK_SIZE);
    if (writePages(fw, dummyPage, 1, curOffset) != 0) {
      free(dummyPage);
      return -1;
    }
  }
  long fwFileSize = getFileSize(fw);
  long ffPagesOffset = zeroPagesOffset + (PAYLOAD_BLOCK_SIZE * (PAYLOAD_TAIL_ZEROPAGE_COUNT + PAYLOAD_TAIL_PADDINGPAGE_COUNT));
  memset(dummyPage, 0xff, sizeof(char) * PAYLOAD_BLOCK_BODY_SIZE);
  while (ffPagesOffset + PAYLOAD_BLOCK_SIZE < fwFileSize) {
    if (writePages(fw, dummyPage, 1, ffPagesOffset) != 0) {
      free(dummyPage);
      return -1;
    }
    ffPagesOffset += PAYLOAD_BLOCK_SIZE;
  }
  free(dummyPage);
  return 0;
}


